#include "topc.h"
#include <stdlib.h> // for atol() [ Note long is only 32 bits on many mach's]

/********************************************************************
 ********************************************************************
 ** GENERAL VERSION:  topcc [ -seq | -mpi | -pthread ] THIS_FILE
 ** ALGORITHM:  Euclidean sieve for factorization
 ** shared data:  num_to_factor (divided by each factor found)
 ** task input:   input (test factor for num_to_factor)
 ** task output:  1 or 0 (true or false)
 ** update:  Divide num_to_factor by num
 ********************************************************************
 ********************************************************************/


#ifndef TASK_INTERVAL
#define TASK_INTERVAL 1000
#endif
/* do AGGLOM_COUNT tasks at once, save on network bandwidth */
#ifndef AGGLOM_COUNT
#define AGGLOM_COUNT 1
#endif


void trace_input( void *input ) { printf("%ld", *(long *)input); } 
void trace_output( void *output ) { if (*(long *)output == 1) printf("TRUE");
                                    if (*(long *)output == 0) printf("FALSE");
                                  }
// This is equivalent to:  long task_out;
// But for SMP (shared memory), it also ensures that task_out is not shared.
#define task_out TOPC_private_global
TOPC_declare_type_of_private_global(long);  // This will be task output

long factors[1000], max_factor=1, num_to_factor, last_num;
void MasSlaveFactor(long);

int main(int argc, char *argv[])
{
  int i;

  TOPC_init(&argc, &argv, TOPC_OPTION("trace",1), // 1 is default
                          TOPC_OPTION("trace_input",trace_input),
                          TOPC_OPTION("trace_output",trace_output),
                          TOPC_OPTION("agglom_count",AGGLOM_COUNT) );
  // Important:  Inspect command line only after TOPC_init()
  if (argc == 1 ) {
    printf("Missing number arg on command line.\n");
    exit(1); }

  MasSlaveFactor( atol(argv[1]) );
  if ( TOPC_is_master() ) {
    for (i = 1; i <= factors[ 0 ]; i++) printf("%ld ", factors[i]);
    printf("\n"); }
  TOPC_finalize();
  // printf("%d: Exiting\n", TOPC_rank()); fflush(stdout);
  exit(0);
}

TOPC_BUF GenerateTaskInput() {
  last_num = last_num + TASK_INTERVAL;
  if ( last_num > num_to_factor ) { return NOTASK; }
  return TOPC_MSG(&last_num, sizeof(last_num));
}

TOPC_BUF DoTask( void *input )
{
  long limit;
  long num = *(long *)input;

  limit = num + TASK_INTERVAL;
  for ( ; num < limit; num++ )
    if (num_to_factor % num == 0) {
      task_out = 1;
      return TOPC_MSG(&task_out, sizeof(long));
    }
  task_out = 0;
  return TOPC_MSG(&task_out, sizeof(long));
}

TOPC_ACTION CheckTaskResult( void *input, void *output )
{
  if ( *(long*)output == 0 ) return NO_ACTION;
  if ( *(long*)input < max_factor ) return UPDATE;
  if ( ! TOPC_is_up_to_date() ) return REDO;
  return UPDATE;
}

void UpdateSharedData( void *input, void *output )
{
  long limit, num = *(long *)input;
  int i;
  limit = num + TASK_INTERVAL;
  // NOTE: Suppose we factor 12, 4 is reported as factor before 2 is reported
  for ( ; num < limit; num++ ) {
    if ( num_to_factor % num == 0 ) {
      if ( num < max_factor ) {
        max_factor = 1; // re-compute max_factor
        for( i = 1; i <= factors[0]; i++ ) {
          if (factors[i] > max_factor) max_factor = factors[i];
          while (factors[i] % num == 0) {
            factors[i] = factors[i] / num;
            num_to_factor = num_to_factor * num;
          }
        }
      }
      while (num_to_factor % num == 0) {
        factors[ ++factors[0] ] = num;
        num_to_factor = num_to_factor / num;
        if (num_to_factor == 1) return;
      }
    }
  }
}

void MasSlaveFactor( long num )
{
  factors[0]=0;
  last_num = 2 - TASK_INTERVAL;
  num_to_factor = num;
  TOPC_master_slave(GenerateTaskInput, DoTask, CheckTaskResult,
                    UpdateSharedData);
}
