 /**********************************************************************
  * TOP-C (Task Oriented Parallel C)                                   *
  * Copyright (c) 2000 Gene Cooperman <gene@ccs.neu.edu>               *
  *                    and Victor Grinberg <victor@ccs.neu.edu>        *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact Gene Cooperman          *
  * <gene@ccs.neu.edu>.                                                *
  **********************************************************************/

/***********************************************************************
 * This is included in topc.c and comm-xxx.c and page.c .  It defines the
 * interface between the two.
 */

#ifndef _COMM_H
#define _COMM_H

#include <stdio.h>  // Needed to define size_t

//Compile-time parameters
#define TOPC_MAX_SLAVES            1000       /* Max. number of slaves */

/***********************************************************************
 * TOP-C data types (shared with topc.h)
 */
#ifndef TOPC_TYPES
#define TOPC_TYPES
typedef int TOPC_BOOL;
typedef struct {
  void  *data;
  size_t data_size;
} TOPC_BUF;
typedef void TOPC_FUNCTION();
struct TOPC_OPT {
       char *name; 
       union { int integer; TOPC_FUNCTION *fnc; } value;
};
#endif

/***********************************************************************
 * TOP-C data types (not in topc.h)
 */
enum TAG {
  TASK_INPUT_TAG,
  REDO_INPUT_TAG,
  CONTINUATION_INPUT_TAG,
  UPDATE_INPUT_TAG,
  UPDATE_OUTPUT_TAG,
  SLAVE_REPLY_TAG,
  END_MASTER_SLAVE_TAG,
  PING_TAG,
  CHDIR_TAG,
  SLAVE_WAIT_TAG,
  PAGE_ACCESS_TAG
};

struct opt_list {
  int count;
  struct TOPC_OPT options[20];
};

extern TOPC_BUF (*COMM_do_task)(void *input); // for comm-seq.c
extern void (*COMM_slave_loop)(void);         // for comm-pthreads.c

/***********************************************************************
 * COMM layer functions (interface between comm-xxx.c and topc.c)
 */
void COMM_init(int *argc, char ***argv, struct opt_list *options_list);
void COMM_finalize(void);
TOPC_BOOL COMM_initialized(void);
int COMM_is_shared_memory(void);
int COMM_node_count(void);
int COMM_rank(void);
TOPC_BOOL COMM_probe_msg(void);
void COMM_send_msg(void *msg, size_t msg_size, int dst, enum TAG tag);
void COMM_receive_msg(void **msg, size_t *msg_size, int *src, enum TAG *tag);
void COMM_free_receive_buf(void *msg);

/***********************************************************************
 * SMP functions
 */
void COMM_begin_atomic_read(void);
void COMM_end_atomic_read(void);
void COMM_begin_atomic_write(void);
void COMM_end_atomic_write(void);
TOPC_BOOL COMM_is_in_atomic_section(void);

// end TOPC_H
#endif
